/*

    Boolean Expression Solver
    Version 2.0

    Module eval.c
    This module contains functions that perform a logical formula
    evaluation based on a passed syntactic tree, i.e. determining
    boolean values for each possible combination of input variables.

    Dialect : ANSI C
    Compiler: any ANSI C-compatible

    Copyright (c) Kamil Ekstein, 2008
    Seminar work of `Programming in C'

    This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
    INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. Kamil Ekstein reserves all rights not expressly
    granted herein.

*/

#include <stdlib.h>
#include "tree.h"
#include "parser.h"
#include "eval.h"
#include "defs.h"

/* ____________________________________________________________________________

    unsigned long int bin_pow(int a)
    
    Returns `a'-th power of two within the range of unsigned long int,
    i.e. 2^a for a in <0; sizeof(unsigned long int) * 8 - 1>.
   ____________________________________________________________________________
*/
unsigned long int bin_pow(int a) {
    int i = 1;
    
    if (a < 0 || a > (sizeof(unsigned long int) * 8 - 1)) return 0;
    else return i << a;
}

/* ____________________________________________________________________________

    void clear_var_values(variable_t var_table[])

    Clears the variables filed in the variable table passed as the
    argument. Called before the evaluation starts to initialize the
    table for the process.
   ____________________________________________________________________________
*/
void clear_var_values(variable_t var_table[]) {
    int i;
    
    for (i = 0; i < MAX_VARTABLE_LENGTH; i++)
        var_table[i].value = 0;
}

/* ____________________________________________________________________________

    int get_var_ones(variable_t var_table[])
    
    Returns a number of variables filed in the variable table passed
    as the argument, that are currently set to logical one.
   ____________________________________________________________________________
*/
int get_var_ones(variable_t var_table[]) {
    int i;
    int c = 0;

    for (i = 0; i < var_index; i++)
        if (var_table[i].value == 1) c++;
        
    return c;
}

/* ____________________________________________________________________________

    void inc_var_values(variable_t var_table[])

    Changes the combination of logical ones and zeros in the boolean
    variables filed in the variable table passed as the argument, so
    that they (when ordered to an integer) compose a value larger by
    one compared to the current state.
   ____________________________________________________________________________
*/
void inc_var_values(variable_t var_table[]) {
    int i = 0;
    int cy = 1;	/* carry flag: set to one when a carry to higher order is performed */
    
    while (cy && i < var_index) {
        if (var_table[i].value == 0) {
            var_table[i].value = 1;
            cy = 0;	/* the carry ends here, zero turned to one */
        }
        else 
            var_table[i].value = 0;

        i++;
    }
}

/* ____________________________________________________________________________

    LOCAL int bin_operation(char operation, int left, int right)

    Performs a boolean operation given by the `operation' argument. Takes
    two integers - `left' and `right' - as the operands of the performed
    operation. Returns the result of the operation.
   ____________________________________________________________________________
*/
LOCAL int bin_operation(char operation, int left, int right) {
    switch (operation) {
        case OP_NEGATION:
            return !left;
        case OP_CONJUNCTION:
            return left && right;
        case OP_DISJUNCTION:
            return left || right;
        case OP_EXCL_DISJUNCTION:
            return left ^ right;
        case OP_IMPLICATION:
            return !left || right;
        case OP_EQUIVALENCE:
            return left == right;
    }

	return 0;
}

/* ____________________________________________________________________________

    int evaluate(node_t *node, variable_t var_table[])

    Recursively evaluates the passed parse tree node using the boolean
    values filed in the passed variable table. Returns a boolean value
    of the given node according to values its sub-trees.
   ____________________________________________________________________________
*/
int evaluate(node_t *node, variable_t var_table[]) {
    int result = 0;
    int lres, rres = 0;
    
    if (node == NULL) return 0;
    
    if (node->left != NULL) lres = evaluate(node->left, var_table);
    if (node->right != NULL) rres = evaluate(node->right, var_table);
    
    if (node->operation == OP_HOLDS_VAR) {
		/* the node is just a sole variable - no unary/binary operation,
		   thus its value must be returned */
        result = var_table[node->variable].value;
    }
    else {
		/* the node represents an operation, thus it must be evaluated
		   and the result is returned */
        result = bin_operation(node->operation, lres, rres);
    }
    
    return result;
}
