/*

    Boolean Expression Solver
    Version 2.0

    Module main.c
    This module composes all the parts of the analyzer together and
    contains the executable entry point function. It is also responsible
    for the user interaction.
    
    Dialect : ANSI C
    Compiler: any ANSI C-compatible

    Copyright (c) Kamil Ekstein, 2008
    Seminar work of `Programming in C'

    This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
    INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. Kamil Ekstein reserves all rights not expressly
    granted herein.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "defs.h"
#include "tree.h"
#include "scan.h"
#include "eval.h"
#include "parser.h"

/* ____________________________________________________________________________
 
    Global Module Variables
   ____________________________________________________________________________
*/
node_t *tree = NULL;                        /* root of the parse tree */
int verb_flag = 0;                          /* verbose flag, indicates whether to print stats */
char dot_fname[MAX_INPUT_LENGTH] = {0};     /* filename of the prospective Graphviz dot file */
FILE *fout = NULL;                          /* output file handle */

/* ____________________________________________________________________________

    void header()
    
    Prints a programme header to the console - namely the programme name
    and a copyright and version information.
   ____________________________________________________________________________
*/
void header() {
    printf("Boolean Expression Solver Version 2.0\n");
    printf("Seminar work of \"Programming in C\"\n");
    printf("Copyright (c) Kamil Ekstein, 2008\n\n");
}

/* ____________________________________________________________________________

    void help()
    
    Prints a short help. Shows how to use the programme, what are the
    command-line options, etc.
   ____________________________________________________________________________
*/
void help() {
    printf("Usage:\n   boole \"<expression>\" [-v] [-g<filename>]\n\n");
    printf("Options:\n   -v ............. Verbose mode (prints out progress messages)\n");
    printf("   -g<filename> ... Generates a Graphviz dot file with the parse tree\n\n");
    printf("Example:\n   boole \"~(a & b) == (~a | ~b)\" -v -gdemorgan.dot\n\n");
    printf("Operators are:\n");
    printf("   ~ .... unary negation (not)\n");
    printf("   & .... conjunction (and)\n");
    printf("   | .... disjunction (or)\n");
    printf("   ^ .... exclusive disjunction (xor)\n");    
    printf("   => ... implication (if-then)\n");    
    printf("   -> ... implication (if-then)\n");    
    printf("   == ... equivalence (if-and-only-if)\n");    
}

/* ____________________________________________________________________________

    void handle_input(int argc, char *argv[])
    
    Parses a command-line input to the programme. If no arguments are
    given, prints out the help. Otherwise processes the arguments and
    sets the respective flags and variables accordingly.
   ____________________________________________________________________________
*/
void handle_input(int argc, char *argv[]) {
    int i;
    
    if (argc < 2) {
        header();
        help();
        exit(1);
    }
    
    strncpy(input, argv[1], MAX_INPUT_LENGTH - 1);
    input[MAX_INPUT_LENGTH - 1] = 0; /* attach the terminating #0 to the end of the string */
    
    for (i = 2; i < argc; i++) {
        if (argv[i][0] != '-') continue;
        switch (toupper(argv[i][1])) {
            case 'V':
                verb_flag = 1;
                header();
                printf("* Verbose mode enabled ...\n");
                break;
            case 'G':
                strncpy(dot_fname, &argv[i][2], MAX_INPUT_LENGTH - 1);
                if (verb_flag)
                    printf("* Generating dot file \"%s\"\n", dot_fname);
                break;
            default:
                printf("Unknown option. Ignoring ...\n");
        }
    }

    if (verb_flag) printf("* Solving expression %s ...\n\n", input);
}

/* ____________________________________________________________________________

    void print_msg(char msg[])
    
    Prints out the passed message string into an output file (given by a
    global variable). If the number of variables in the expression is
    upto 5, the message is printed also to the console.
   ____________________________________________________________________________
*/
void print_msg(char msg[]) {
    if (var_index <= 5) printf(msg);
    fprintf(fout, msg);
}

/* ____________________________________________________________________________

    void print_table_header()
    
    Prints out the result table header and grid using the print_msg calls.
   ____________________________________________________________________________ 
*/
void print_table_header() {
    char msg[COM_STRING_LENGTH] = {0};
    int i, j;

    /* print the result table header */
    for (i = 0; i < var_index; i++) {
        sprintf(msg, " %s |", var_table[i].name);
        print_msg(msg);
    }
    print_msg(" $\n");

    /* print the result table grid */
    for (i = 0; i < var_index; i++) {
        print_msg("-");
        for (j = 0; j < strlen(var_table[i].name); j++)
            print_msg("-");
        print_msg("-+");
    }
    print_msg("---\n");
}

/* ____________________________________________________________________________

    void print_vars()
    
    Prints out the result table rows, i.e. the values of the variables
    and the overall result of the expression evaluation for the given
    values of the variables.
   ____________________________________________________________________________
*/
void print_vars() {
    char msg[COM_STRING_LENGTH] = {0};    
    int i, j;
    
    for (i = 0; i < var_index; i++) {
        print_msg(" ");
        for (j = 0; j < strlen(var_table[i].name) - 1; j++)
            print_msg(" ");
        sprintf(msg, "%d |", var_table[i].value);
        print_msg(msg);
    }
}

/* ____________________________________________________________________________

    int compute()

    Performs the whole computation of the boolean formula to be
    evaluated. Returns an indicator whether the formula is a
    contradiction, tautology, or none of the above.
   ____________________________________________________________________________
*/
int compute() {
    int r, i;
    int all_ones = 1, all_zeros = 1;	/* flags indicating tautology or contradiction */
    int num_rows = 0;					/* number of rows in the resulting table */
    char msg[COM_STRING_LENGTH] = {0};
    
    clear_var_values(var_table);
    print_table_header();

    num_rows = bin_pow(var_index);		/* get the number of rows and 2^(number of vars) */
    for (i = 0; i < num_rows; i++) {
        r = evaluate(tree, var_table);	/* evaluate the parse tree for all possible combinations of values */
        
        if (r == 0) all_ones = 0;		/* update the tautology/contradiction flags */
        else all_zeros = 0;
        
        print_vars();
        sprintf(msg, " %d\n", r);
        print_msg(msg);
        
        inc_var_values(var_table);        
    }
    
    if (all_zeros && all_ones) return (int) 'E';    /* error, this cannot happen any time */
    if (all_zeros) return (int) 'C';                /* contradiction */
    if (all_ones) return (int) 'T';                 /* tautology */
    return (int) 'N';                               /* non-specific */
}

/* ____________________________________________________________________________

    void shutdown()
    
    Shuts the programme down correctly, i.e. disposes dynamic variables
    and closes open files, etc.
   ____________________________________________________________________________
*/
void shutdown() {
    dispose_node(&tree);
    if (fout != NULL) fclose(fout);
    printf("\n");
}

/* ____________________________________________________________________________

    void setup(int argc, char *argv[])
    
    Starts the programme, i.e. registers exit procedures, checks the passed
    command-line arguments, sets global variables, allocates memory, etc.
   ____________________________________________________________________________
*/
void setup(int argc, char *argv[]) {
    char fname[MAX_INPUT_LENGTH] = {0};
    time_t time_stamp = 0;
    struct tm *now = NULL;
    int reg_c;

    /* register an exit procedure to cleanup the stuff on dirty aborts */
    reg_c = atexit(shutdown);
    assert(reg_c == 0);

	/* parse command-line arguments */
    handle_input(argc, argv);

	/* set global variables and open files */
    time_stamp = time(NULL);
    now = localtime(&time_stamp);
    strftime(fname, MAX_INPUT_LENGTH, "%Y-%m-%d-%H%M%S.bool", now);
    fout = fopen(fname, "w");
}

/* ____________________________________________________________________________

    void run()
    
    The main executive routine - performs the whole computation. Used
    to put all the parts of the evaluation together in a compact way,
    not in the main function.
   ____________________________________________________________________________
*/
void run() {
    char msg[MAX_INPUT_LENGTH] = {0};
    
	/* produce a parse tree by parsing the input expression */
    tree = parse();
    
	/* check the parse tree validity */
    if (tree != NULL) {
        sprintf(msg, "PARSER-RSL=OK; VARS=%d; D$=%ld;\n\n",
            var_index, bin_pow(var_index));
        print_msg(msg);
    }

    /* if a dot file name is given, produce a dot file */
    if (strlen(dot_fname)) dump_node_as_dot(dot_fname, tree);
   
	/* report on the character of the analysed formula */
    switch (compute()) {
        case 'N': print_msg("\nNONSPECIFIC\n"); break;
        case 'C': print_msg("\nCONTRADICTION\n"); break;
        case 'T': print_msg("\nTAUTOLOGY\n"); break;
        default: print_msg("\nERROR\n");
    }
}

/* ____________________________________________________________________________
 
    MAIN PROGRAM
   ____________________________________________________________________________
*/

int main(int argc, char *argv[]) {
    setup(argc, argv);
    run();
    shutdown();
    
    return 0;
}
