/*

    Boolean Expression Solver
    Version 2.0

    Module parser.c
    This module is responsible for a syntactic analysis of input strings.
    Determines syntax validity and parses the boolean formula given as input
    producing a parse tree suitable for a formula evaluation.

    Dialect : ANSI C
    Compiler: any ANSI C-compatible

    Copyright (c) Kamil Ekstein, 2008
    Seminar work of `Programming in C'

    This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
    INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. Kamil Ekstein reserves all rights not expressly
    granted herein.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "parser.h"
#include "tree.h"
#include "scan.h"
#include "err.h"

/* ____________________________________________________________________________

    Global Module Variables
   ____________________________________________________________________________
*/
char msg[MAX_INPUT_LENGTH];                 /* internal global string var used for sprintf() calls */
variable_t var_table[MAX_VARTABLE_LENGTH];  /* variable table */
int var_index = 0;                          /* variable table index */
token_t tok;                                /* currently scanned lexical token from input */
node_t *node = NULL;                        /* internal global pointer to the so far created parse tree */

/* ____________________________________________________________________________

    Function Prototypes
   ____________________________________________________________________________
*/
LOCAL void parse_expression();              /* a function defined after its first use */

/* ____________________________________________________________________________

    LOCAL int file_variable(char *id)
    
    Files a variable named id into the variable table.
    Returns the position of the newly filed variable in the table.
   ____________________________________________________________________________
*/
LOCAL int file_variable(char *id) {
    int i = var_index;
    
    if (var_index >= MAX_VARTABLE_LENGTH)
        raise_error("too many variables - infeasible");

    strncpy(var_table[var_index].name, id, MAX_IDENTIFIER_LENGTH);
    var_table[var_index].value = 0;
    var_index++;

    return i;
}

/* ____________________________________________________________________________

    LOCAL int get_variable(char *id)
    
    Searches for the variable named id in the variable table. If found
    returns its position (index) otherwise -1.
   ____________________________________________________________________________
*/
LOCAL int get_variable(char *id) {
    int i;
    
    for (i = 0; i < var_index; i++)
        if (strncmp(var_table[i].name, id, MAX_IDENTIFIER_LENGTH) == 0) {
            return i;
        }
        
    return -1;
}

/* ____________________________________________________________________________

    char *name_operation(char opcode)
    
    Transforms an opcode given by the opcode argument into its lexical
    representation, i.e. returns a string with a printable name of the
    operation.
   ____________________________________________________________________________
*/
char *name_operation(char opcode) {
    switch (opcode) {
        case OP_NEGATION:
            return "NOT";
        case OP_CONJUNCTION:
            return "AND";
        case OP_DISJUNCTION:
            return "OR";
        case OP_EXCL_DISJUNCTION:
            return "XOR";
        case OP_IMPLICATION:
            return "IMP";
        case OP_EQUIVALENCE:
            return "EQU";
        case OP_HOLDS_VAR:
            return "VAR";
        case OP_HOLDS_NEG_VAR:
            return "~VAR";
        case OP_HOLDS_TRUE:
            return "CONST_1";
        case OP_HOLDS_FALSE:
            return "CONST_0";
        default:
            return "UNKNOWN";
    }
}

/* ____________________________________________________________________________

    LOCAL int match(token_t *tok, int expected_tok_code)
    
    Checks whether the current input token given by the tok argument matches
    the expected (according to the grammar) token expected_tok_code. Returns
    the following token code or TOK_UNKNOWN if an unexpected token found.
   ____________________________________________________________________________
*/
LOCAL int match(token_t *tok, int expected_tok_code) {
    char msg[COM_STRING_LENGTH];
    
    if (tok->code != expected_tok_code) {
        sprintf(msg, "%s expected at position %d",
            name_token(expected_tok_code), inp_index);
        raise_error(msg);
        return TOK_UNKNOWN;
    }
    else {
        scan(tok);                          /* if the token was expected, continue scanning the input */
        return tok->code;
    }
}

/* ____________________________________________________________________________

    LOCAL void parse_atom()
    
    Part of the recursive descent parser: Parses atomic parts of the formula,
    i.e. constants, variables and expressions in parentheses. Input taken
    via scan() and match() function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_atom() {
    switch (tok.code) {
        case TOK_TRUE:
            node = create_node(OP_HOLDS_TRUE, NULL, NULL);
            scan(&tok);
            break;
        case TOK_FALSE:
            node = create_node(OP_HOLDS_FALSE, NULL, NULL);
            scan(&tok);
            break;
        case TOK_VAR:
            node = create_node(OP_HOLDS_VAR, NULL, NULL);
            node->variable = get_variable(tok.value);           /* check for existence of the variable */
            if (node->variable == -1)                           /* if not found ... */
                node->variable = file_variable(tok.value);      /* file the new variable into the variable table */
            scan(&tok);
            break;
        case TOK_LPAR:
            match(&tok, TOK_LPAR);                              /* if an open parenthesis found */
            parse_expression();                                 /* try to parse a bracketed expression */
            match(&tok, TOK_RPAR);                              /* and check for the closing parenthesis */
            break;
        default:
            sprintf(msg, "missing const, var or '(' at position %d", inp_index);
            raise_error(msg);
    }
}

/* ____________________________________________________________________________

    LOCAL void parse_term()

    Part of the recursive descent parser: Parses superatomic parts of the
    formula, i.e. unary operation(s). Input taken via scan() and match()
    function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_term() {
    node_t *child = NULL;

    if (tok.code == TOK_NEGATION) {
        scan(&tok);
        parse_atom();
        child = node;
        node = create_node(OP_NEGATION, child, NULL);
    }
    else parse_atom();
}

/* ____________________________________________________________________________

    LOCAL void parse_conjunct()

    Part of the recursive descent parser: Parses conjunct expressions of
    the type "X & Y". Input taken via scan() and match() function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_conjunct() {
    node_t *lchild, *rchild = NULL;

    parse_term();
    
    while (tok.code == TOK_CONJUNCTION) {
        lchild = node;        
        scan(&tok);

        parse_term();
        rchild = node;
    
        if (lchild != NULL && rchild != NULL)                       /* if both left and right part of the expression found */
            node = create_node(OP_CONJUNCTION, lchild, rchild);     /* create a new superordinate node in the parse tree */
    }
}

/* ____________________________________________________________________________

    LOCAL void parse_disjunct()

    Part of the recursive descent parser: Parses disjunct expressions of
    the type "X | Y" and "X ^ Y" (xor). Input taken via scan() and match()
    function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_disjunct() {
    node_t *lchild, *rchild = NULL;
    int op;

    parse_conjunct();

    while (tok.code == TOK_DISJUNCTION || tok.code == TOK_EXCL_DISJUNCTION) {
        lchild = node;
        op = tok.code;
        scan(&tok);
        
        parse_conjunct();
        rchild = node;

        if (lchild != NULL && rchild != NULL)                       /* if both left and right part of the expression found */
            switch (op) {                                           /* create a new superordinate node in the parse tree ... */
                case TOK_DISJUNCTION:
                    node = create_node(OP_DISJUNCTION, lchild, rchild);
                    break;
                case TOK_EXCL_DISJUNCTION:
                    node = create_node(OP_EXCL_DISJUNCTION, lchild, rchild);
            }
    }
}

/* ____________________________________________________________________________

    LOCAL void parse_implicat()

    Part of the recursive descent parser: Parses implication expressions of
    the type "X -> Y". Input taken via scan() and match() function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_implicat() {
    node_t *lchild, *rchild = NULL;
    
    parse_disjunct();
    
    while (tok.code == TOK_IMPLICATION) {
        lchild = node;
        scan(&tok);
        
        parse_disjunct();
        rchild = node;

        if (lchild != NULL && rchild != NULL)                       /* if both left and right part of the expression found */
            node = create_node(OP_IMPLICATION, lchild, rchild);     /* create a new superordinate node in the parse tree */
    }
}

/* ____________________________________________________________________________

    LOCAL void parse_expression()

    Part of the recursive descent parser: Parses equivalence expressions of
    the type "X == Y". Input taken via scan() and match() function calls.
   ____________________________________________________________________________
*/
LOCAL void parse_expression() {
    node_t *lchild, *rchild = NULL;
    
    parse_implicat();

    while (tok.code == TOK_EQUIVALENCE) {
        lchild = node;        
        scan(&tok);
        
        parse_implicat();
        rchild = node;
        
        if (lchild != NULL && rchild != NULL)                       /* if both left and right part of the expression found */
            node = create_node(OP_EQUIVALENCE, lchild, rchild);     /* create a new superordinate node in the parse tree */
    }
}

/* ____________________________________________________________________________

    node_t *parse()
    
    Parses the input string. Input taken via scan(), match() and recursive
    descent parser function calls.
    Returns a top-most node (root) of the newly created parse tree.
   ____________________________________________________________________________
*/
node_t *parse() {
    /* Clean up the parser structures */
    if (node != NULL) dispose_node(&node);
    memset(var_table, 0, sizeof(var_table));
    
	/* do the parsing */
    scan(&tok);
    parse_expression();
    
    if (!fully_analyzed()) raise_error("syntax error");
        
    return node;
}
