/*

    Boolean Expression Solver
    Version 2.0

    Module scan.c
    This module contains a lexical analyser (or scanner) and its support
	stuff. The scanner analyses	an input string and splits it into lexical
	atoms fed into the syntactic analyser (or parser).

    Dialect : ANSI C
    Compiler: any ANSI C-compatible

    Copyright (c) Kamil Ekstein, 2008
    Seminar work of `Programming in C'

    This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
    INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. Kamil Ekstein reserves all rights not expressly
    granted herein.

*/

#include <ctype.h>
#include <string.h>
#include "defs.h"
#include "scan.h"

/* ____________________________________________________________________________

    Global Module Variables
   ____________________________________________________________________________
*/
char input[MAX_INPUT_LENGTH] = {0};	/* buffer for an input string */
int inp_index = 0;					/* pointer to the currently analysed input char */
int look_ahead = 0;					/* flag indicating whether a one char ahead was read */
int current;						/* value of the currently analysed char */

/* ____________________________________________________________________________

    int fully_analyzed()
    
    Returns a non-zero value in the case that the input string was entirely
    analysed, i.e. the input index reached the end of the input string.
   ____________________________________________________________________________
*/
int fully_analyzed() {
    return (inp_index == strlen(input));
}

/* ____________________________________________________________________________

    LOCAL int get_next()
    
    Provides the calling routine with a next character from the input string
    and moves the input index one char further.
   ____________________________________________________________________________
*/
LOCAL int get_next() {
    int c;
    
	/* check for the end of the input string */
    if (inp_index >= strlen(input)) return LAT_STOPPER;

    c = toupper((int) input[inp_index]);	/* case INsensitive: all chars capitalised */
    inp_index++;
    
    return c;
}

/* ____________________________________________________________________________

    char *name_token(int tok_code)
    
    Returns a printable name of a lexical token given by the `tok_code'
    argument. Used for debugging purposes.
   ____________________________________________________________________________
*/
char *name_token(int tok_code) {
    switch (tok_code) {
        case TOK_FALSE:
            return "CONST_0";
        case TOK_TRUE:
            return "CONST_1";
        case TOK_NEGATION:
            return "'~' (NOT)";
        case TOK_CONJUNCTION:
            return "'&' (AND)";
        case TOK_DISJUNCTION:
            return "'|' (OR)";
        case TOK_EXCL_DISJUNCTION:
            return "'^' (XOR)";
        case TOK_IMPLICATION:
            return "'->' (IMP)";
        case TOK_EQUIVALENCE:
            return "'==' (EQU)";
        case TOK_LPAR:
            return "'(' (LPAR)";
        case TOK_RPAR:
            return "')' (RPAR)";
        case TOK_VAR:
            return "VAR";
        case TOK_STOPPER:
            return "STOP.";
        default:
            return "UNKNOWN";
    }
}

/* ____________________________________________________________________________

    void scan(token_t *token)
    
    Fills a variable given by the passed `token' pointer with a lexical
    token extracted from the input string. While extracting the token, all
    blank chars are skipped.
   ____________________________________________________________________________
*/
void scan(token_t *token) {
    int v_idx = 0;
    
	/* do not read a new char if it has been pre-read */
    if (look_ahead) look_ahead = 0;
    else current = get_next();
    
    /* Skip the spaces, tabs, etc. */
    while (isspace(current)) current = get_next();
    
    switch (current) {
        case LAT_STOPPER:
            token->code = TOK_STOPPER;
            break;
        case '0':
            token->code = TOK_FALSE;
            break;
        case '1':
            token->code = TOK_TRUE;
            break;
        case '~':
            token->code = TOK_NEGATION;
            break;
        case '&':
            token->code = TOK_CONJUNCTION;
            break;
        case '|':
            token->code = TOK_DISJUNCTION;
            break;
        case '^':
            token->code = TOK_EXCL_DISJUNCTION;
            break;
        case '-':
            current = get_next();
            if (current == '>') token->code = TOK_IMPLICATION;
            else {
                token->code = TOK_UNKNOWN;                
                look_ahead = 1;
            }
            break;
        case '=':
            current = get_next();
            switch (current) {
                case '=':
                    token->code = TOK_EQUIVALENCE;
                    break;
                case '>':
                    token->code = TOK_IMPLICATION;
                    break;
                default:
                    token->code = TOK_UNKNOWN;
                    look_ahead = 1;
            }
            break;
        case '(':
            token->code = TOK_LPAR;
            break;
        case ')':
            token->code = TOK_RPAR;
            break;
        default:
			/* verify if the atom is an identifier */
            if (isalpha(current)) {
                do {
                    token->value[v_idx] = current;
                    current = get_next();
                    v_idx++;
                } while (isalpha(current) || isdigit(current));

                token->value[v_idx] = 0;    /* terminate the recognized identifier */
                token->code = TOK_VAR;
                look_ahead = 1;
            }
            else token->code = TOK_UNKNOWN;
    }
}
