/*

    Boolean Expression Solver
    Version 2.0

    Module tree.c
    This module contains a set of functions that assures a maintenance of
	binary trees used in the programme to store the parser results, i.e.
	a structure of the analysed boolean formula.

    Dialect : ANSI C
    Compiler: any ANSI C-compatible

    Copyright (c) Kamil Ekstein, 2008
    Seminar work of `Programming in C'

    This file is provided "AS IS" with NO WARRANTY OF ANY KIND,
    INCLUDING THE WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. Kamil Ekstein reserves all rights not expressly
    granted herein.

*/

#include <stdlib.h>
#include <stdio.h>
#include "defs.h"
#include "parser.h"
#include "tree.h"
#include "scan.h"

/* ____________________________________________________________________________

    node_t *create_node(char operation, node_t *left, node_t *right)
    
    Creates a new node at a top of the parse tree setting its operation
    member variable to `operation' and attaches `left' and `right' sub-
    trees passed as arguments as left and right offsprings. Returns
    a pointer to the newly created node.
   ____________________________________________________________________________
*/
node_t *create_node(char operation, node_t *left, node_t *right) {
    node_t *temp = NULL;

    temp = (node_t *) malloc(sizeof(node_t));
    temp->operation = operation;
    temp->left = left;
    temp->right = right;
    temp->variable = 0;
    
    return temp;
}

/* ____________________________________________________________________________

    void dispose_node(node_t **node)
    
    Frees the memory occupied by a tree given by the passed `node' argument.
   ____________________________________________________________________________
*/
void dispose_node(node_t **node) {
    if (*node == NULL) return;
    
    if ((*node)->left != NULL) dispose_node(&(*node)->left);
    if ((*node)->right != NULL) dispose_node(&(*node)->right);

    free(*node);
    *node = NULL;
}

/* ____________________________________________________________________________

    LOCAL void dump_node_rp(FILE *f, node_t *node, int level)
    
    Recursively prints out the contents of a node given by the `node' argument
    to the passed file `f'. The output is a source code for the GraphViz dot
    programme that performs a device-independent tree drawing.
   ____________________________________________________________________________
*/
LOCAL void dump_node_rp(FILE *f, node_t *node, int level) {
    int dlev, dlevp;
    
    if (node == NULL) return;
    
	/* assign all the tree nodes a unique id computed from
	   their position in the tree */
    dlev = level * 2;
    dlevp = dlev + 1;

	/* generate a code for nodes */
    if (node->operation == OP_HOLDS_VAR ||
        node->operation == OP_HOLDS_NEG_VAR)
        fprintf(f, "  %d [label=\"%s\\n%s\", style=filled,"
            " color=\"#FFD700\", fontname=\"Helvetica\", fontsize=\"12\"];\n",
            level, name_operation(node->operation),
            var_table[node->variable].name);
    else
        fprintf(f, "  %d [shape=box, label=\"%s\", style=filled,"
            " color=\"#ADFF2F\", fontname=\"Helvetica\"];\n",
            level, name_operation(node->operation));

	/* generate a code for links */
    if (node->left != NULL) {
        fprintf(f, "  %d -> %d;\n", level, dlev);
        dump_node_rp(f, node->left, dlev);        
    }
    if (node->right != NULL) {
        fprintf(f, "  %d -> %d;\n", level, dlevp);
        dump_node_rp(f, node->right, dlevp);
    }
}

/* ____________________________________________________________________________

    void dump_node_as_dot(char *filename, node_t *node)
    
    Puts out a tree given by its root `node' into the file named `filename'
    by means of generating a source code for the GraphViz dot programme that
    performs a device-independent tree drawing.
   ____________________________________________________________________________
*/
void dump_node_as_dot(char *filename, node_t *node) {
    FILE *fout;
    
    fout = fopen(filename, "w");
    if (fout == NULL) return;

    /* Print out a Graphviz dot header */
    fprintf(fout, "/* Graph of a tree generated by Boole */\n");
    fprintf(fout, "digraph N {\n");

    /* Recursive print out the tree */    
    dump_node_rp(fout, node, 1);

    /* Print out a Graphviz dot footer */
    fprintf(fout, "}\n");    
    
    fclose(fout);
}
